#include "ptexlib.h"
#include "pkin.h"
#include <kpathsea/tex-glyph.h>
#include <kpathsea/magstep.h>

#define T3_BUF_SIZE   1024

#define T3_TYPE_PK    0
#define T3_TYPE_PGC   1

static char t3_line[T3_BUF_SIZE], *t3_line_ptr;
FILE *t3_file;
static boolean t3_image_used;

static unsigned short t3_char_procs[MAX_CHAR_CODE + 1];
static unsigned short t3_char_widths[MAX_CHAR_CODE + 1];
static unsigned short t3_glyph_num;
static float t3_font_scale;
static long int t3_font_type;
static long int t3_b0, t3_b1, t3_b2, t3_b3;

#define t3_open()       \
    open_input(&t3_file, kpse_miscfonts_format, FOPEN_RBIN_MODE)
#define t3_close()      xfclose(t3_file, cur_file_name)
#define t3_getchar()    xgetc(t3_file)
#define t3_eof()        feof(t3_file)
#define t3_prefix(s)    (!strncmp(t3_line, s, strlen(s)))
#define t3_putchar(c)   pdfout(c)

#define t3_check_eof()                                     \
    if (t3_eof())                                          \
        pdftex_fail("unexpected end of file");

static void t3_getline(void) 
{
    int c;
restart:
    t3_line_ptr = t3_line;
    c = t3_getchar();
    while (!t3_eof()) {
        append_char_to_buf(c, t3_line_ptr, t3_line, T3_BUF_SIZE);
        if (c == 10)
            break;
        c = t3_getchar();
    }
    append_eol(t3_line_ptr, t3_line, T3_BUF_SIZE);
    if (t3_line_ptr - t3_line <= 1 || *t3_line == '%') {
        if (!t3_eof())
            goto restart;
    }
}

static void t3_putline(void)
{
    char *p = t3_line;
    while (p < t3_line_ptr)
        t3_putchar(*p++);
}

/*
static void t3_print_char_width(internalfontnumber f, integer c)
{
    pdfprintint(dividescaled(getcharwidth(f, c),
                             xnoverd(pdffontsize[f], 1000, t3_font_scale), 3));
}
*/

static void update_bbox(integer llx, integer lly, integer urx, integer ury)
{
    if (t3_glyph_num == 0) {
        t3_b0 = llx;
        t3_b1 = lly;
        t3_b2 = urx;
        t3_b3 = ury;
    }
    else {
        if (llx < t3_b0)
            t3_b0 = llx;
        if (lly < t3_b1)
            t3_b1 = lly;
        if (urx > t3_b2)
            t3_b2 = urx;
        if (ury > t3_b3)
            t3_b3 = ury;
    }
}

static void t3_write_glyph(internalfontnumber f)
{
    static char t3_begin_glyph_str[] = "\\pdfglyph";
    static char t3_end_glyph_str[] = "\\endglyph";
    int glyph_index;
    int width, height, depth, llx, lly, urx, ury;
    char *p;
    t3_getline();
    if (t3_prefix(t3_begin_glyph_str)) {
        if (sscanf(t3_line + strlen(t3_begin_glyph_str) + 1,
                   "%i %i %i %i %i %i %i %i =", &glyph_index,
                   &width, &height, &depth, &llx, &lly, &urx, &ury) != 8) {
            remove_eol(p, t3_line);
            pdftex_warn("invalid glyph preamble: `%s'", t3_line);
            return;
        }
        if (glyph_index < fontbc[f] || glyph_index > fontec[f])
            return;
    }
    else
        return;
    if (!pdfcharmarked(f, glyph_index)) {
        while (!t3_prefix(t3_end_glyph_str)) {
            t3_check_eof();
            t3_getline();
        }
        return;
    }
    update_bbox(llx, lly, urx, ury);
    t3_glyph_num++;
    pdfnewdict(0, 0);
    t3_char_procs[glyph_index] = objptr;
    if (width == 0) 
        t3_char_widths[glyph_index] = t3_font_scale*
            ((float)getcharwidth(f, glyph_index)/pdffontsize[f]);
    else
        t3_char_widths[glyph_index] = width;
    pdfbeginstream();
    t3_getline();
    pdf_printf("%i 0 %i %i %i %i d1\nq\n", 
               (int) t3_char_widths[glyph_index], (int)llx,
               (int)lly, (int)urx, (int)ury);
    while (!t3_prefix(t3_end_glyph_str)) {
        t3_check_eof();
        if (t3_prefix("BI"))
            t3_image_used = true;
        t3_putline();
        t3_getline();
    }
    pdf_puts("Q\n");
    pdfendstream();
}

static boolean writepk(internalfontnumber f)
{
    kpse_glyph_file_type font_ret;
    integer res = cfgvalue(cfgpkresolutioncode);
    integer llx, lly, urx, ury;
    integer cw, rw, i, j;
    halfword *row;
    float width_scale;
    char *name;
    chardesc cd;
    integer dpi;
    if (res == 0)
        res = 600;
    dpi = 
        kpse_magstep_fix((unsigned)(res*((float)fontsize[f]/fontdsize[f])+0.5),
                         res, NULL);
    cur_file_name =  makecstring(fontname[f]);
    name = kpse_find_pk(cur_file_name, dpi, &font_ret);
    if (name == 0 ||
        !FILESTRCASEEQ(cur_file_name, font_ret.name) ||
        !kpse_bitmap_tolerance((double)font_ret.dpi, (double) dpi)) {
        pdftex_warn("Font %s at %i not found", cur_file_name, (int)dpi);
        cur_file_name = 0;
        return false;
    }
    t3_font_type = T3_TYPE_PK;
    t3_image_used = true;
    t3_font_scale = (res/72.0)*((float)pdffontsize[f]/pdf1bp);
    width_scale = res/(72.0*pdf1bp);
    t3_file = xfopen(name, FOPEN_RBIN_MODE);
    tex_printf(" <%s", (char *)name);
    while (readchar((t3_glyph_num == 0), &cd) != 0) {
        t3_glyph_num++;
        if (!pdfcharmarked(f, cd.charcode) ||
            cd.cwidth < 1 || cd.cheight < 1) {
            xfree(cd.raster);
            continue;
        }
        llx = -cd.xoff;
        lly = cd.yoff - cd.cheight + 1;
        urx = cd.cwidth + llx + 1;
        ury = cd.cheight + lly;
        update_bbox(llx, lly, urx, ury);
        pdfnewdict(0, 0);
        t3_char_procs[cd.charcode] = objptr;
        t3_char_widths[cd.charcode] = getcharwidth(f, cd.charcode)*width_scale;
        pdfbeginstream();
        pdf_printf("%i 0 %i %i %i %i d1\nq\n", (int)cd.xescape, 
                   (int)llx, (int)lly, 
                   (int)urx, (int)ury);
        pdf_printf("%i 0 0 %i %i %i cm\nBI\n", (int)cd.cwidth,
                   (int)cd.cheight, (int)llx, (int)lly);
        pdf_printf("/W %i\n/H %i\n", 
                   (int)cd.cwidth, (int)cd.cheight);
        pdf_puts("/IM true\n/BPC 1\n/D [1 0]\nID ");
        cw = (cd.cwidth + 7)/8;
        rw = (cd.cwidth + 15)/16;
        row = cd.raster;
        for (i = 0; i < cd.cheight; i++) {
            for (j = 0; j < rw - 1; j++) {
                pdfout(*row/256);
                pdfout(*row%256);
                row++;
            }
            pdfout(*row/256);
            if (2*rw == cw)
                pdfout(*row%256);
            row++;
        }
        xfree(cd.raster);
        pdf_puts("\nEI\nQ\n");
        pdfendstream();
    }
    cur_file_name = 0;
    return true;
}

void writet3(int objnum, internalfontnumber f)
{
    static char t3_font_scale_str[] = "\\pdffontscale";
    int i;
    t3_glyph_num = 0;
    t3_font_type = T3_TYPE_PGC;
    t3_image_used = false;
    for (i = 0; i <= MAX_CHAR_CODE; i++) {
        t3_char_procs[i] = 0;
        t3_char_widths[i] = 0;
    }
    packfilename(fontname[f], getnullstr(), maketexstring(".pgc"));
    cur_file_name = makecstring(makenamestring());
    if (!t3_open()) {
        if (writepk(f))
            goto write_font_dict;
        else {
            cur_file_name = 0;
            return;
        }
    }
    tex_printf("<%s", nameoffile + 1);
    t3_getline();
    if (!t3_prefix(t3_font_scale_str) ||
        sscanf(t3_line + strlen(t3_font_scale_str) + 1, "%f", &t3_font_scale) < 1 ||
        t3_font_scale <= 0 || t3_font_scale > 1000 ) {
        pdftex_warn("missing or invalid font scale");
        t3_close();
        cur_file_name = 0;
        return;
    }
    while (!t3_eof())
        t3_write_glyph(f);
write_font_dict:
    pdfbegindict(objnum); /* Type 3 font dictionary */
    pdf_puts("/Type /Font\n/Subtype /Type3\n");
    pdf_printf("/Name /F%i\n", (int)f);
    pdf_printf("/FontMatrix [%.5g 0 0 %.5g 0 0]\n", 
               1/t3_font_scale, 1/t3_font_scale);
    pdf_printf("/%s [ %i %i %i %i ]\n", 
               font_keys[FONTBBOX1_CODE].pdfname, 
               (int)t3_b0, (int)t3_b1, 
               (int)t3_b2, (int)t3_b3);
    pdf_printf("/Resources << /ProcSet [ /PDF %s] >>\n", 
               t3_image_used ? "/ImageB " : "");
    pdf_printf("/FirstChar %i\n/LastChar %i\n", 
               (int)fontbc[f], (int)fontec[f]);
    pdf_printf("/Widths %i 0 R\n/Encoding %i 0 R\n/CharProcs %i 0 R\n", 
               (int)(objptr + 1), (int)(objptr + 2),
               (int)(objptr + 3));
    pdfenddict();
    pdfnewobj(0, 0); /* chars width array */
    pdf_puts("[");
    for (i = fontbc[f]; i <= fontec[f]; i++)
        pdf_printf("%i ", (int)t3_char_widths[i]);
    pdf_puts("]\n");
    pdfendobj();
    pdfnewdict(0, 0); /* encoding dictionary */
    pdf_printf("/Type /Encoding\n/Differences [%i ", (int)fontbc[f]);
    for (i = fontbc[f]; i <= fontec[f]; i++)
        pdf_printf("/a%i", (int)i);
    pdf_puts("]\n");
    pdfenddict();
    pdfnewdict(0, 0); /* CharProcs dictionary */
    for (i = fontbc[f]; i <= fontec[f]; i++)
        if (t3_char_procs[i] != 0)
            pdf_printf("/a%i %i 0 R\n", 
                       (int)i, (int)t3_char_procs[i]);
    pdfenddict();
    t3_close();
    tex_printf(">");
    cur_file_name = 0;
}
